#include "SVNClientPackerExtension.h"
#include "JNIUtil.h"
#include "Pool.h"
#include <svn_path.h>
#include "Revision.h"
#include "../include/org_tigris_subversion_javahl_Revision.h"

SVNClientPackerExtension::SVNClientPackerExtension(SVNClient* super)
  : m_super(super)
{
}

SVNClientPackerExtension::~SVNClientPackerExtension()
{
}

SVNClientPackerExtension * SVNClientPackerExtension::getCppObject(jobject jthis)
{
    static jfieldID fid = 0;
    JNIEnv *env = JNIUtil::getEnv();
    if(fid == 0)
    {
        jclass clazz = env->FindClass(JAVA_PACKAGE"/SVNClientForPacker");
        if(JNIUtil::isJavaExceptionThrown())
        {
            return NULL;
        }
        fid = env->GetFieldID(clazz, "cppAddr", "J");
        if(JNIUtil::isJavaExceptionThrown())
        {
            return NULL;
        }
    }

    jlong cppAddr = env->GetLongField(jthis, fid);
    if(JNIUtil::isJavaExceptionThrown())
    {
        return NULL;
    }
    return reinterpret_cast<SVNClientPackerExtension*>(cppAddr);
}

jlong SVNClientPackerExtension::getCppAddr()
{
    return reinterpret_cast<jlong>(this);
}


void SVNClientPackerExtension::dispose(jobject jthis)
{
    delete this;
    static jfieldID fid = 0;
    JNIEnv *env = JNIUtil::getEnv();
    if(fid == 0)
    {
        jclass clazz = env->FindClass(JAVA_PACKAGE"/SVNClientForPacker");
        if(JNIUtil::isJavaExceptionThrown())
        {
            return;
        }
        fid = env->GetFieldID(clazz, "cppAddr", "J");
        if(JNIUtil::isJavaExceptionThrown())
        {
            return;
        }
    }

    env->SetLongField(jthis, fid, 0);
    if(JNIUtil::isJavaExceptionThrown())
    {
        return;
    }
}

void SVNClientPackerExtension::finalize()
{
    JNIUtil::putFinalizedClient(this);
}

jobject SVNClientPackerExtension::headRevision(jobject jthis, const char *path)
{
  Pool subPool;
  apr_pool_t * apr_pool = subPool.pool ();

  m_super->m_lastPath = svn_path_internal_style (path, apr_pool);
	
  svn_client_ctx_t *ctx = m_super->getContext(NULL);
  if(ctx == NULL)
  {
	return NULL;
  }
  const char *URL;

  svn_error_t * error = svn_client_url_from_path (&URL, path, apr_pool);


  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return NULL;
  }
   
  if(URL == NULL)
  {
	  JNIUtil::handleSVNError(svn_error_create(SVN_ERR_UNVERSIONED_RESOURCE, NULL,
                                "Either a URL or versioned item is required."));
	  return NULL;
  }
  
  svn_revnum_t set_rev;
  error = svn_client_head_revision(URL,&set_rev, ctx, apr_pool);

  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return NULL;
  }

  return createJavaRevision(jthis,set_rev);
}

// [dh]
jstring SVNClientPackerExtension::repositoryUrl(jobject jthis, const char *path)
{
  Pool subPool;
  apr_pool_t * apr_pool = subPool.pool ();
  m_super->m_lastPath = svn_path_internal_style (path, apr_pool);
	
  svn_client_ctx_t *ctx = m_super->getContext(NULL);
  if(ctx == NULL)
  {
	return NULL;
  }

  svn_stringbuf_t* repos_root;
  svn_error_t * error = svn_client_repos_root (path, &repos_root, ctx, apr_pool);

  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return NULL;
  }

  jstring jReposRoot = JNIUtil::makeJString(repos_root->data);
  if(JNIUtil::isJavaExceptionThrown())
  {
      return NULL;
  }
  
  return jReposRoot;
}

void SVNClientPackerExtension::rename(const char *srcPath, const char *destPath, 
                     const char *message, Revision &revision, bool force)
{
    Pool requestPool;
    apr_pool_t * apr_pool = requestPool.pool ();

    if(srcPath == NULL)
    {
        JNIUtil::throwNullPointerException("srcPath");
        return;
    }
    if(destPath == NULL)
    {
        JNIUtil::throwNullPointerException("destPath");
        return;
    }
    svn_client_commit_info_t *commit_info = NULL;
    Path sourcePath = srcPath;
    m_super->m_lastPath = svn_path_internal_style (destPath, apr_pool);
    svn_client_ctx_t *ctx = m_super->getContext(message);
    if(ctx == NULL)
    {
        return;
    }


    svn_error_t *err = svn_client_rename (&commit_info,
                                        sourcePath.c_str (),
                                        revision.revision (),
                                        m_super->m_lastPath.c_str (),
                                        force,
                                        ctx,
                                        apr_pool);
    if(err != NULL)
         JNIUtil::handleSVNError(err);
}

jlong SVNClientPackerExtension::put(jobject jthis, const char* srcPath, const char* destPath,  bool forbidOverwrite, const char* message)
{
  Pool subPool;
  apr_pool_t * apr_pool = subPool.pool ();
  
  if(srcPath == NULL)
    {
      JNIUtil::throwNullPointerException("srcPath");
      return 0;
    }
  if(destPath == NULL)
    {
      JNIUtil::throwNullPointerException("destPath");
      return 0;
    }

  Path sourcePath = srcPath;
  m_super->m_lastPath = svn_path_internal_style (destPath, apr_pool);
  svn_client_commit_info_t *commit_info = NULL;
  
  svn_client_ctx_t *ctx = m_super->getContext(message);
  if(ctx == NULL)
  {
	return 0;
  }

  svn_error_t * error = svn_client_put(&commit_info, sourcePath.c_str (), m_super->m_lastPath.c_str(), forbidOverwrite, ctx, apr_pool);

  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return 0;
  }

  if(commit_info && SVN_IS_VALID_REVNUM (commit_info->revision))
    {
      return commit_info->revision;
    }
  else
    {
      return -1;
    }
}

jlong SVNClientPackerExtension::propertySet(jobject jthis, const char* url, const char* propName, const char* propVal, const char* message)
{
  Pool subPool;
  apr_pool_t * apr_pool = subPool.pool ();
  
  if(url == NULL)
    {
      JNIUtil::throwNullPointerException("url");
      return 0;
    }
  if(propName == NULL)
    {
      JNIUtil::throwNullPointerException("propName");
      return 0;
    }

   m_super->m_lastPath = svn_path_internal_style (url, apr_pool);
 
  svn_client_commit_info_t *commit_info = NULL;
  
  svn_client_ctx_t *ctx = m_super->getContext(message);

  if(ctx == NULL)
  {
	return 0;
  }
  
  svn_string_t* prop_val;
  if (propVal == NULL) {
    prop_val = NULL;
  }
  else {
    prop_val = svn_string_create(propVal,apr_pool);
  }

  svn_error_t * error = svn_client_prop_set(&commit_info,  propName, prop_val,
					    m_super->m_lastPath.c_str(),
					    ctx, apr_pool);
 

  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return 0;
  }

  if(commit_info && SVN_IS_VALID_REVNUM (commit_info->revision))
    {
      return commit_info->revision;
    }
  else
    {
      return -1;
    }
}

jobjectArray SVNClientPackerExtension::properties(jobject jthis, const char *path, Revision &revision) {
    apr_array_header_t * props;
    Pool requestPool;
    if(path == NULL)
    {
        JNIUtil::throwNullPointerException("path");
        return NULL;
    }
    apr_pool_t * apr_pool = requestPool.pool ();
    m_super->m_lastPath = svn_path_internal_style (path, apr_pool);

    svn_client_ctx_t *ctx = m_super->getContext(NULL);
    if(ctx == NULL)
    {
        return NULL;
    }

    svn_error_t *Err = svn_client_proplist (&props,
                                            m_super->m_lastPath.c_str (),
                                            revision.revision(),
                                            FALSE,
                                            ctx,
                                            apr_pool);
    if(Err != NULL)
    {
        JNIUtil::handleSVNError(Err);
        return NULL;
    }

    // since we disabled recurse, props->nelts should be 1
    for (int j = 0; j < props->nelts; ++j)
    {
        svn_client_proplist_item_t *item =
            ((svn_client_proplist_item_t **)props->elts)[j];

        apr_hash_index_t *hi;

        int count = apr_hash_count (item->prop_hash);

        JNIEnv *env = JNIUtil::getEnv();
        jclass clazz = env->FindClass(JAVA_PACKAGE"/PropertyData");
        if(JNIUtil::isJavaExceptionThrown())
        {
            return NULL;
        }
        jobjectArray ret = env->NewObjectArray(count, clazz, NULL);
        if(JNIUtil::isJavaExceptionThrown())
        {
            return NULL;
        }
        env->DeleteLocalRef(clazz);
        if(JNIUtil::isJavaExceptionThrown())
        {
            return NULL;
        }

        int i = 0;
        for (hi = apr_hash_first (apr_pool, item->prop_hash); hi;
             hi = apr_hash_next (hi), i++)
        {
            const char *key;
            svn_string_t *val;

            apr_hash_this (hi, (const void **)&key, NULL, (void**)&val);

            jobject object = m_super->createJavaProperty(jthis, item->node_name->data, 
                                                key, val);

            env->SetObjectArrayElement(ret, i, object);
            if(JNIUtil::isJavaExceptionThrown())
            {
                return NULL;
            }
            env->DeleteLocalRef(object);
            if(JNIUtil::isJavaExceptionThrown())
            {
                return NULL;
            }
        }
        return ret;
    }
    return NULL;
}

jlong SVNClientPackerExtension::propertyRemove(jobject jthis, const char* url, const char* propName, const char* message)
{
  return propertySet(jthis,url,propName,NULL,message);
}

typedef struct
{
  svn_node_kind_t kind;
} kind_baton_t;

static void
report_kind_callback(void* baton,
		     const char* path,
		     svn_node_kind_t kind)
{
  kind_baton_t* kb = (kind_baton_t*)baton;

  kb->kind = kind;
}

jint SVNClientPackerExtension::kind(const char *path, Revision &revision)
{
  Pool subPool;
  apr_pool_t * apr_pool = subPool.pool ();

  kind_baton_t kind_baton;

  if(path == NULL)
    {
      JNIUtil::throwNullPointerException("path");
      return 0;
    }
  m_super->m_lastPath = svn_path_internal_style (path, apr_pool);
  
  svn_client_ctx_t * ctx = m_super->getContext(NULL);
  if(ctx == NULL)
    {
      return 0;
    }

  svn_error_t *error = svn_client_kind(m_super->m_lastPath.c_str(), revision.revision(),
				     report_kind_callback,
				     &kind_baton,
				     ctx,apr_pool);


  if(error != SVN_NO_ERROR)
  {
 	JNIUtil::handleSVNError(error);
	return 0;
  }
  return kind_baton.kind;
}

jobject SVNClientPackerExtension::createJavaRevision(jobject jthis, svn_revnum_t rev_num)
{
	JNIEnv *env = JNIUtil::getEnv();
	jclass clazz = env->FindClass(JAVA_PACKAGE"/Revision$Number");
	if(JNIUtil::isJavaExceptionThrown())
	{
		return NULL;
	}
	static jmethodID mid = 0;
	if(mid == 0)
	{
		mid = env->GetMethodID(clazz, "<init>", "(J)V");
		if(JNIUtil::isJavaExceptionThrown())
		{
			return NULL;
		}
	}

        jlong jRevNum = rev_num;

	jobject ret = env->NewObject(clazz, mid, jRevNum);
	if(JNIUtil::isJavaExceptionThrown())
	{
		return NULL;
	}
	env->DeleteLocalRef(clazz);
	if(JNIUtil::isJavaExceptionThrown())
	{
		return NULL;
	}
	return ret;
}
