#define USE_DEBUG_DEFAULTS

/*** Includes. ***/

#include "svn_wc.h"
#include "svn_client.h"
#include "svn_string.h"
#include "svn_path.h"
#include "svn_delta.h"
#include "svn_error.h"
#include "cl.h"
#include "svn_private_config.h"

#include "svn_packer_extensions.h"


/*** Code. ***/

/* This implements the `svn_opt_subcommand_t' interface. */
svn_error_t *
svn_cl__put (apr_getopt_t *os,
	     void *baton,
	     apr_pool_t *pool)
{
#ifdef USE_DEBUG_DEFAULTS
  static const char* debug_src_path = "a_file";
  static const char* debug_dst_url = "http://dhoffmann:50080/packer/repos/test/a_file";
#endif

  svn_cl__opt_state_t *opt_state = ((svn_cl__cmd_baton_t *) baton)->opt_state;
  svn_client_ctx_t *ctx = ((svn_cl__cmd_baton_t *) baton)->ctx;
  apr_array_header_t *targets;
  const char *src_path;
  const char *dst_url;
  svn_client_commit_info_t *commit_info = NULL;

  SVN_ERR (svn_opt_args_to_target_array (&targets, os, 
                                         opt_state->targets,
                                         &(opt_state->start_revision),
                                         &(opt_state->end_revision),
                                         FALSE, pool));

  if (targets->nelts != 2) {
#ifdef USE_DEBUG_DEFAULTS
    src_path = debug_src_path;
    dst_url = debug_dst_url;
#else
    return svn_error_create (SVN_ERR_CL_ARG_PARSING_ERROR, 0, NULL);
#endif
  }
  else {
    src_path = ((const char **) (targets->elts))[0];
    dst_url = ((const char **) (targets->elts))[1];
  }

  if (! svn_path_is_url (dst_url))
    return svn_error_createf
      (SVN_ERR_CL_ARG_PARSING_ERROR, NULL,
       _("Invalid URL '%s'"), dst_url);

  if (! opt_state->quiet)
    svn_cl__get_notifier (&ctx->notify_func, &ctx->notify_baton,
                          FALSE, FALSE, FALSE, pool);

  SVN_ERR (svn_cl__make_log_msg_baton (&(ctx->log_msg_baton), opt_state, 
                                       NULL, ctx->config, pool));

  SVN_ERR (svn_cl__cleanup_log_msg 
           (ctx->log_msg_baton, svn_client_put (&commit_info,
						src_path,
						dst_url,
						FALSE,
						ctx,
						pool)));

  if (commit_info && ! opt_state->quiet)
    SVN_ERR (svn_cl__print_commit_info (commit_info, pool));

  return SVN_NO_ERROR;
}
