#include "svn_private_config.h"
#include "svn_pools.h"
#include "svn_packer_extensions.h"
#include "svn_error.h"
#include "svn_path.h"
#include "client.h"
#include "svn_md5.h"

#define USE_SEND_FILE_CONTENTS_FROM_IMPORT
#define USE_CHECK_PATH

#ifdef USE_SEND_FILE_CONTENTS_FROM_IMPORT
#include "svn_subst.h"
#endif

/*
  examples taken from:
  subversion
    +- libsvn_client
        +-copy.c
	+-add.c
*/


/*** Code. ***/

#ifdef USE_SEND_FILE_CONTENTS_FROM_IMPORT
static svn_error_t *
send_file_contents (const char *path,
                    void *file_baton,
                    const svn_delta_editor_t *editor,
                    apr_hash_t *properties,
                    unsigned char *digest,
                    apr_pool_t *pool)
{
  const char *tmpfile_path = NULL;
  svn_stream_t *contents;
  svn_txdelta_window_handler_t handler;
  void *handler_baton;
  apr_file_t *f = NULL;
  svn_error_t *err = SVN_NO_ERROR;
  const svn_string_t *eol_style_val = NULL, *keywords_val = NULL;
  svn_boolean_t special = FALSE;

  /* If there are properties, look for EOL-style and keywords ones. */
  if (properties)
    {
      eol_style_val = apr_hash_get (properties, SVN_PROP_EOL_STYLE,
                                    sizeof (SVN_PROP_EOL_STYLE) - 1);
      keywords_val = apr_hash_get (properties, SVN_PROP_KEYWORDS,
                                   sizeof (SVN_PROP_KEYWORDS) - 1);
      if (apr_hash_get (properties, SVN_PROP_SPECIAL, APR_HASH_KEY_STRING))
        special = TRUE;
    }

  /* Get an editor func that wants to consume the delta stream. */
  SVN_ERR (editor->apply_textdelta (file_baton, NULL, pool,
                                    &handler, &handler_baton));

  /* If we have EOL styles or keywords to de-translate, do it.  Any
     EOL style gets translated to "\n", the repository-normal format.
     Keywords get unexpanded.  */
  if (eol_style_val || keywords_val || special)
    {
      svn_subst_keywords_t keywords = {0};
      const char *temp_dir;
      apr_file_t *tmp_f;

      /* Now create a new tempfile, and open a stream to it. */
      SVN_ERR (svn_io_temp_dir (&temp_dir, pool));
      SVN_ERR (svn_io_open_unique_file 
               (&tmp_f, &tmpfile_path,
                svn_path_join (temp_dir, "svn-import", pool),
                ".tmp", FALSE, pool));
      SVN_ERR (svn_io_file_close (tmp_f, pool));

      /* Generate a keyword structure. */
      if (keywords_val)
        SVN_ERR (svn_subst_build_keywords (&keywords, keywords_val->data, 
                                           APR_STRINGIFY(SVN_INVALID_REVNUM),
                                           "", 0, "", pool));
      
      if ((err = svn_subst_copy_and_translate2 (path, tmpfile_path,
                                                eol_style_val ? "\n" : NULL,
                                                FALSE,
                                                keywords_val ? &keywords : NULL,
                                                FALSE,
                                                special,
                                                pool)))
        goto cleanup;
    }

  /* Open our contents file, either the original path or the temporary
     copy we might have made above. */
  if ((err = svn_io_file_open (&f, tmpfile_path ? tmpfile_path : path, 
                               APR_READ, APR_OS_DEFAULT, pool)))
    goto cleanup;
  contents = svn_stream_from_aprfile (f, pool);

  /* Send the file's contents to the delta-window handler. */
  if ((err = svn_txdelta_send_stream (contents, handler, handler_baton,
                                      digest, pool)))
    goto cleanup;

  /* Close our contents file. */
  if ((err = svn_io_file_close (f, pool)))
    goto cleanup;

 cleanup:
  if (tmpfile_path)
    {
      /* If we used a tempfile, we need to close and remove it, too. */
      svn_error_t *err2 = svn_io_remove_file (tmpfile_path, pool);
      if (err && err2)
        svn_error_compose (err, err2);
      else if (err2)
        err = err2;
    }

  return err;
}
#else
static svn_error_t *
send_file_contents(const char* src_path,
		   void* file_baton,
		   const svn_delta_editor_t *editor,
		   apr_hash_t *properties,
		   unsigned char *digest,
		   apr_pool_t *pool)
{
  void* handler_baton = NULL;
  svn_txdelta_window_handler_t handler;
  apr_file_t* src_file = NULL;
  
  svn_stream_t *src_stream = NULL;
    
  SVN_ERR (editor->apply_textdelta (file_baton, NULL, pool, &handler, &handler_baton));
  
  SVN_ERR (svn_io_file_open(&src_file, src_path, APR_READ, APR_OS_DEFAULT, pool));
  
  src_stream = svn_stream_from_aprfile (src_file, pool);

  SVN_ERR (svn_txdelta_send_stream(src_stream, handler, handler_baton, digest, pool));

  SVN_ERR (svn_io_file_close(src_file, pool));

  return SVN_NO_ERROR;
}
#endif

/* Operations between get_commit_editor() and close_edit()/abort_edit() are
   put into a separate function.
*/
static svn_error_t *
put(svn_client_ctx_t *ctx, const svn_delta_editor_t *editor, void* edit_baton,
    const char * src_path,
    const char *parent_url, const char *child_name, svn_boolean_t file_exists, svn_revnum_t rev,
    apr_pool_t *pool)
{
  void* root_baton = NULL;
  unsigned char digest[APR_MD5_DIGESTSIZE];
  const char *text_checksum;
  void* file_baton = NULL;
  apr_hash_t* properties = NULL;
  apr_hash_index_t *hi;

  svn_node_kind_t kind;
  svn_boolean_t is_special;
  const char *mimetype = NULL;

  SVN_ERR (editor->open_root(edit_baton, rev, pool, &root_baton));

  /* Add the new version of the file */

  if (file_exists) {
    SVN_ERR (editor->open_file(child_name,root_baton, rev, pool, &file_baton));
  }
  else {
    SVN_ERR (editor->add_file(child_name,root_baton, NULL, SVN_INVALID_REVNUM, pool, &file_baton));
  }

  SVN_ERR (svn_io_check_special_path (src_path, &kind, &is_special, pool));

  if (is_special) {
    properties = apr_hash_make (pool);
  }
  else {
    SVN_ERR (svn_client__get_auto_props (&properties, &mimetype, src_path, ctx,
					 pool));
  }

  if (properties) {
    for (hi = apr_hash_first (pool, properties); hi; hi = apr_hash_next (hi)) {
      const void *pname;
      void *pval;
      
      apr_hash_this (hi, &pname, NULL, &pval);
      SVN_ERR (editor->change_file_prop (file_baton, pname, pval, pool));
    }
  }


  /* If this is a special file, we need to set the svn:special
     property and create a temporary detranslated version in order to
     send to the server. */
  if (is_special)
    {
      apr_hash_set (properties, SVN_PROP_SPECIAL, APR_HASH_KEY_STRING,
                    svn_string_create (SVN_PROP_SPECIAL_VALUE, pool));
      SVN_ERR (editor->change_file_prop (file_baton, SVN_PROP_SPECIAL,
                                         apr_hash_get (properties,
                                                       SVN_PROP_SPECIAL,
                                                       APR_HASH_KEY_STRING),
                                         pool));
    }

  SVN_ERR (send_file_contents(src_path,file_baton,editor,
			      properties,digest,pool));

  text_checksum = svn_md5_digest_to_cstring (digest, pool);
  SVN_ERR (editor->close_file(file_baton, text_checksum, pool));

  return SVN_NO_ERROR;
}

static svn_error_t *
set_prop(const svn_delta_editor_t *editor, void* edit_baton, const char* file_name, svn_revnum_t rev,
	 const char* name, const svn_string_t* value,  apr_pool_t *pool)
{
  void* root_baton = NULL;
  void* file_baton = NULL;

  SVN_ERR (editor->open_root(edit_baton, rev, pool, &root_baton));
  SVN_ERR (editor->open_file(file_name, root_baton, rev, pool, &file_baton));
  SVN_ERR (editor->change_file_prop (file_baton, name, (void*)value, pool));
  SVN_ERR (editor->close_file(file_baton, NULL, pool));

  return SVN_NO_ERROR;
}


svn_error_t *
svn_client_head_revision(const char *URL,
			 svn_revnum_t *revnum,
			 svn_client_ctx_t *ctx,
			 apr_pool_t *pool)
{
  void *ra_baton, *session;
  svn_ra_plugin_t *ra_lib;

  /* Open an RA session for the URL. Note that we don't have a local
     directory, nor a place to put temp files. */
  SVN_ERR (svn_ra_init_ra_libs (&ra_baton, pool));
  SVN_ERR (svn_ra_get_ra_library (&ra_lib, ra_baton, URL, pool));
  SVN_ERR (svn_client__open_ra_session (&session, ra_lib, URL, NULL,
                                        NULL, NULL, FALSE, TRUE,
                                        ctx, pool));
  
  SVN_ERR (ra_lib->get_latest_revnum (session, revnum, pool));
  return SVN_NO_ERROR;
}

svn_error_t *
svn_client_repos_root(const char *URL,
		      svn_stringbuf_t ** root_url,
		      svn_client_ctx_t *ctx,
		      apr_pool_t *pool)
{
  void *ra_baton, *session;
  svn_ra_plugin_t *ra_lib;

  /* Open an RA session for the URL. Note that we don't have a local
     directory, nor a place to put temp files. */
  SVN_ERR (svn_ra_init_ra_libs (&ra_baton, pool));
  SVN_ERR (svn_ra_get_ra_library (&ra_lib, ra_baton, URL, pool));
  SVN_ERR (svn_client__open_ra_session (&session, ra_lib, URL, NULL,
                                        NULL, NULL, FALSE, TRUE,
                                        ctx, pool));

  const char* root_url_cstr;
  SVN_ERR (ra_lib->get_repos_root (session, &root_url_cstr, pool));
  (*root_url) = svn_stringbuf_create(root_url_cstr,pool);
  return SVN_NO_ERROR;
}

svn_error_t *
svn_client_put(svn_client_commit_info_t **commit_info,
	       const char *src_path,
	       const char *dst_url,
	       svn_boolean_t forbid_overwrite,
	       svn_client_ctx_t *ctx,
	       apr_pool_t *pool)
{
  const char *parent_url, *child_name, *message;
  void *ra_baton, *session;
  svn_ra_plugin_t *ra_lib;
  const svn_delta_editor_t *editor;
  void *edit_baton;
  
  void *commit_baton = NULL;

  const char *abs_src_path;
  const char *tmp_file;

  svn_boolean_t file_exists;
  svn_error_t* err;

  svn_revnum_t head_rev;

  svn_node_kind_t kind;

#ifndef USE_CHECK_PATH
  apr_hash_t *dirents;
  svn_dirent_t *dirent;
#endif

  /* The commit process uses absolute paths, so we need to open the access
     baton using absolute paths, and so we really need to use absolute
     paths everywhere. */
  SVN_ERR (svn_path_get_absolute (&abs_src_path, src_path, pool));
  
  /* Split the DST_URL into an parent_url and child_name. */
  svn_path_split (dst_url, &parent_url, &child_name, pool);

  /* Get the RA vtable that matches URL. */
  SVN_ERR (svn_ra_init_ra_libs (&ra_baton, pool));
  SVN_ERR (svn_ra_get_ra_library (&ra_lib, ra_baton, parent_url, pool));

  SVN_ERR (svn_client__open_ra_session (&session, ra_lib, parent_url, NULL,
                                        NULL, NULL, FALSE, TRUE,
                                        ctx, pool));

  SVN_ERR (ra_lib->get_latest_revnum (session, &head_rev, pool));

#ifdef USE_CHECK_PATH 
  SVN_ERR (ra_lib->check_path (session, child_name, head_rev, &kind, pool));
  file_exists = (kind != svn_node_none);
#else
  SVN_ERR (ra_lib->get_dir (session, "", head_rev, &dirents, NULL, NULL, pool));

  dirent = apr_hash_get (dirents, child_name, APR_HASH_KEY_STRING);
  file_exists = (dirent != NULL);
  kind = dirent->kind;
#endif
  if (file_exists) {
    if (kind != svn_node_file) {
      return svn_error_createf(SVN_ERR_CLIENT_IS_DIRECTORY, NULL,
			      _("Cannot overwrite non file object %s in repository"),dst_url);
    }
    else if (forbid_overwrite) {
      return svn_error_createf(SVN_ERR_FS_ALREADY_EXISTS, NULL,
                                _("File '%s' already exists"),dst_url);
    }
  }

  svn_client_commit_item_t *item;
  apr_array_header_t *commit_items 
    = apr_array_make (pool, 1, sizeof (item));

  item = apr_pcalloc (pool, sizeof (*item));
  item->url = apr_pstrdup (pool, dst_url);
  if (file_exists) {
    item->state_flags = SVN_CLIENT_COMMIT_ITEM_TEXT_MODS;
  }
  else {
    item->state_flags = SVN_CLIENT_COMMIT_ITEM_ADD;
  }

  (*((svn_client_commit_item_t **) apr_array_push (commit_items))) = item;
  
  /* Get the commit message with whatever callback function and baton
     is provided through the context */
  SVN_ERR ((*ctx->log_msg_func) (&message, &tmp_file, commit_items, 
				 ctx->log_msg_baton, pool));

  /* Create a new commit baton and initialize it with pool and commit_info 
   * struct commit_baton is defined along svn_client__commit_get_baton()
   * and svn_client__commit_callback(). The latter will fill in values in
   * commit_info.
   */
  SVN_ERR (svn_client__commit_get_baton (&commit_baton, commit_info, pool));

  SVN_ERR (ra_lib->get_commit_editor (session, &editor, &edit_baton, message,
                                      svn_client__commit_callback, /* called on close_edit */
                                      commit_baton, pool));
  
  err = put(ctx,editor,edit_baton,
	    abs_src_path,
	    parent_url,child_name,
	    file_exists,head_rev,pool);

  if (err != SVN_NO_ERROR) {
    svn_error_clear (editor->abort_edit (edit_baton, pool));
    return err;
  }

  SVN_ERR (editor->close_edit (edit_baton, pool));

  return SVN_NO_ERROR;	   
}

svn_error_t *
svn_client_kind (const char *path_or_url,
		 const svn_opt_revision_t *revision,
		 svn_report_kind_func_t report_kind_func,
		 void* report_kind_baton,
		 svn_client_ctx_t *ctx,
		 apr_pool_t *pool)
{
  svn_ra_plugin_t *ra_lib;
  void *session;
  svn_revnum_t rev;
  svn_node_kind_t url_kind;
  const char *url;

  /* Get an RA plugin for this filesystem object. */
  SVN_ERR (svn_client__ra_lib_from_path (&ra_lib, &session, &rev,
                                         &url, path_or_url, revision,
                                         ctx, pool));

  SVN_ERR (ra_lib->check_path (session, "", rev, &url_kind, pool));

  report_kind_func(report_kind_baton,path_or_url,url_kind);

  return SVN_NO_ERROR;
}

/* Check whether NAME is a revision property name.
 * 
 * Return TRUE if it is.
 * Return FALSE if it is not.  
 */ 
static svn_boolean_t
is_revision_prop_name (const char *name)
{
  apr_size_t i;
  const char *revision_props[] = 
    {
      SVN_PROP_REVISION_ALL_PROPS
    };

  for (i = 0; i < sizeof (revision_props) / sizeof (revision_props[0]); i++)
    {
      if (strcmp (name, revision_props[i]) == 0)
        return TRUE;
    }
  return FALSE;
}

svn_error_t*
svn_client_prop_set (svn_client_commit_info_t **commit_info,
		     const char *propname,
		     const svn_string_t *propval,
		     const char *target,
		     svn_client_ctx_t *ctx,
		     apr_pool_t *pool)
{
  const char *parent_url, *child_name, *message;
  void *ra_baton, *session;
  svn_ra_plugin_t *ra_lib;
  const svn_delta_editor_t *editor;
  void *edit_baton;
  
  void *commit_baton = NULL;

  const char *tmp_file;

  svn_error_t* err;

  svn_revnum_t head_rev;

  svn_node_kind_t kind;
  
  if (is_revision_prop_name (propname))
    {
      return svn_error_createf (SVN_ERR_CLIENT_PROPERTY_NAME, NULL,
                                _("Revision property '%s' not allowed "
                                  "in this context"), propname);
    }

  if (svn_path_is_url (target))
    {
      /* Split the TARGET into an parent_url and child_name. */
      svn_path_split (target, &parent_url, &child_name, pool);
      
      /* Get the RA vtable that matches URL. */
      SVN_ERR (svn_ra_init_ra_libs (&ra_baton, pool));
      SVN_ERR (svn_ra_get_ra_library (&ra_lib, ra_baton, parent_url, pool));
      
      SVN_ERR (svn_client__open_ra_session (&session, ra_lib, parent_url, NULL,
					    NULL, NULL, FALSE, TRUE,
					    ctx, pool));
      
      SVN_ERR (ra_lib->get_latest_revnum (session, &head_rev, pool));

      SVN_ERR (ra_lib->check_path (session, child_name, head_rev, &kind, pool));

      switch (kind) {
      case svn_node_file:
      case svn_node_dir:
	break;
      default:
	 return svn_error_createf(SVN_ERR_NODE_UNEXPECTED_KIND, NULL,
			      _("Attempt to set property on non-file/non-directory resource %s"),target);
      }

      
      svn_client_commit_item_t *item;
      apr_array_header_t *commit_items 
	= apr_array_make (pool, 1, sizeof (item));
      
      item = apr_pcalloc (pool, sizeof (*item));
      item->url = apr_pstrdup (pool, target);

      item->state_flags = SVN_CLIENT_COMMIT_ITEM_PROP_MODS;

      (*((svn_client_commit_item_t **) apr_array_push (commit_items))) = item;

      /* Get the commit message with whatever callback function and baton
	 is provided through the context */
      SVN_ERR ((*ctx->log_msg_func) (&message, &tmp_file, commit_items, 
				     ctx->log_msg_baton, pool));
      
      /* Create a new commit baton and initialize it with pool and commit_info 
       * struct commit_baton is defined along svn_client__commit_get_baton()
       * and svn_client__commit_callback(). The latter will fill in values in
       * commit_info.
       */
      SVN_ERR (svn_client__commit_get_baton (&commit_baton, commit_info, pool));
      
      SVN_ERR (ra_lib->get_commit_editor (session, &editor, &edit_baton, message,
					  svn_client__commit_callback, /* called on close_edit */
					  commit_baton, pool));
      
      err = set_prop(editor,edit_baton,child_name,head_rev,propname,propval,pool);
      
      if (err != SVN_NO_ERROR) {
	svn_error_clear (editor->abort_edit (edit_baton, pool));
	return err;
      }
      
      SVN_ERR (editor->close_edit (edit_baton, pool));
      
      return SVN_NO_ERROR;	   
    }
  else {
    return svn_client_propset(propname,propval,target,FALSE,pool);
  }
}


svn_error_t*
svn_client_get_delta_editor(svn_client_commit_info_t **commit_info,
			    const char* root_url,
			    svn_client_ctx_t *ctx,
			    const char* message,
			    const svn_delta_editor_t **editor,
			    void** edit_baton,
			    apr_pool_t *pool)
{
  void *ra_baton, *session;
  svn_ra_plugin_t *ra_lib;
  void *commit_baton = NULL;

    /* Get the RA vtable that matches URL. */
  SVN_ERR (svn_ra_init_ra_libs (&ra_baton, pool));
  SVN_ERR (svn_ra_get_ra_library (&ra_lib, ra_baton, root_url, pool));

  SVN_ERR (svn_client__open_ra_session (&session, ra_lib, root_url, NULL,
                                        NULL, NULL, FALSE, TRUE,
                                        ctx, pool));

 
  /* Create a new commit baton and initialize it with pool and commit_info 
   * struct commit_baton is defined along svn_client__commit_get_baton()
   * and svn_client__commit_callback(). The latter will fill in values in
   * commit_info.
   */
  SVN_ERR (svn_client__commit_get_baton (&commit_baton, commit_info, pool));

  SVN_ERR (ra_lib->get_commit_editor (session, editor, edit_baton, message,
                                      svn_client__commit_callback, /* called on close_edit */
                                      commit_baton, pool));

  return SVN_NO_ERROR;
}

